-- Happy Holidays
-- By Tim Sandford, 2019
-- Adapted from the board game originally created by Paul Sandford
-- Comments to timsandford42@hotmail.com
function love.run()  -- Main function
  gVERSION, gPATCH, gREVISION, gCODENAME = love.getVersion()
  if gVERSION >=  11 then  -- Colour fix for 11.0 onwards
    local oldSetColor = love.graphics.setColor
    love.graphics.setColor = function (r, g, b, a)     
      if type(r)=="table" then
        r = r[1] / 255
        g = r[2] / 255
        b = r[3] / 255
        a = (r[4] or 255) / 255
      else
        r = r / 255
        g = g / 255
        b = b / 255
        a = (a or 255) / 255
      end
      oldSetColor(r,g,b,a)
    end
  end
  math.randomseed(os.time())  -- True randomness
  love.load(love.arg.parseGameArguments(arg), arg) -- load variables etc.
  while true do  -- Main loop
    if love.event then -- Process events
      love.event.pump()
      for name, a,b,c,d,e,f in love.event.poll() do
        if name == "quit" then
          if not love.quit or not love.quit() then
            return a
          end
        end
        love.handlers[name](a,b,c,d,e,f)
      end
    end
    love.update() -- Update
    love.graphics.origin() -- Not sure, probably reset origin?
    love.graphics.clear(love.graphics.getBackgroundColor()) -- Clears the screen
    love.draw() -- Collect draw commands
    love.graphics.present() -- Present the graphics on the screen!
    love.timer.sleep(0.001) -- This is to lower the load on the CPU, but I could edit it :)	
  end
end

function love.load(arg)  -- Load everything in
  local i, s
  io.stdout:setvbuf("no")                                         -- Print messages immediately, allows tracing bugs
  if arg[#arg] == "-debug" then require("mobdebug").start() end   -- Zerobrane debugging
  gGAMEMODE = "Title"                                             -- Set gamemode
  gSCREENWIDTH = love.graphics.getWidth()                         -- Get screensize
  gSCREENHEIGHT = love.graphics.getHeight()                       -- Get screensize
  gRATIOX = gSCREENWIDTH / 800                                    -- Scale graphics to screen
  gRATIOY = gSCREENHEIGHT / 600                                   -- Scale graphics to screen
  gFADE = 255                                                     -- Fade in effect
  gFADESPEED = 10                                                 -- Speed of fade in effect
  gFPS = 60                                                       -- Frames per second
  gTIMER = love.timer.getTime()                                   -- Use timer in love.update for speed of game
  gSOUNDON = true                                                 -- Sounds on
  gMUSICON = true                                                 -- Music on
  gFLASH = 255			                                               -- Flashing text transparency
  gFLASH_z = 10 			                                             -- Speed of change for flashing text
  gFLASH_min = 80                                                 -- Minimum transparency of anything flashing
  gFLASH_max = 255                                                -- Maximum transparency of anything flashing
  gTYPING = false                                                 -- Flag to input text
  gTYPEDTEXT = ""                                                 -- Text input
  gSELECT = 0                                                     -- Selected town/city etc.
  gSUBSELECT = 0                                                  -- Selected activity to change
  gSTAGE = 0                                                      -- Turn stage
  gATTEMPTED = false                                              -- Have they attempted an activity that turn?
  gDICE = 0                                                       -- What number did the player roll?
  gDELAY = 0                                                      -- Countdown for all delayed actions
  gTURN = 0                                                       -- Whose turn is it?
  gPLAYERS = 0                                                    -- Number of players
  gZOOMSPEED = 1.06                                               -- Speed of zooming in/out
  gZOOM = 0                                                       -- Zoomed in how much
  gZOOM_MIN, gZOOM_MAX = 0, 0                                     -- Minimum & maximum zoom
  gCAMERAX, gCAMERAY = 0, 0                                       -- Camera coordinates
  gMOUSEHELD = 0                                                  -- How long has the mouse button been held?
  gBRUSH = 0                                                      -- Choose a map brush 1-8
  gMAPNAME = "default"                                            -- When they load a map
  gWEATHER, gNEXT_WEATHER = 0, 0                                  -- Choose weather 1-6
  gHAZARD, gSERVICE, gLOCATION = 0, 0, 0                          -- Choose hazard/service card (some hazards have specific location)
  gPOLICE = "A.I."                                                -- Either Player/A.I./Off
  
  sSOUND = {}                                                     -- Load sounds
  for i = 1, 99, 1 do
    if i < 10 then s = "sounds/sound0"..i..".ogg"
    else s = "sounds/sound"..i..".ogg" end
    if love.filesystem.getInfo(s, "file") then sSOUND[i] = love.audio.newSource(s, "static") end
  end
  
  gMUSIC = love.audio.newSource("music/music.mp3", "stream")
  gMUSIC:setLooping(true)
  gMUSIC:setVolume(0.5)
  
  if love.filesystem.getInfo("gui/cursor.png", "file") then       -- Mouse pointer
    gCURSOR = love.mouse.newCursor("gui/cursor.png", 0, 0)
    love.mouse.setCursor(gCURSOR)
  end
  love.mouse.setVisible(true)
  
  fTINY = love.graphics.newFont("fonts/font.ttf", 10)              -- tiny font
  fSMALL = love.graphics.newFont("fonts/font.ttf", 16)            -- small font
  fDEFAULT = love.graphics.newFont("fonts/font.ttf", 24)          -- default font
  fHEADING = love.graphics.newFont("fonts/font.ttf", 32)          -- heading font
  fTITLE = love.graphics.newFont("fonts/font.ttf", 40)            -- title font
    
  iIMAGE = {}                                                     -- Load images
  for i = 1, 99, 1 do
    if i < 10 then s = "images/image0"..i..".png"
    else s = "images/image"..i..".png" end
    if love.filesystem.getInfo(s, "file") then iIMAGE[i] = love.graphics.newImage(s) end
  end
  
  tTILES = {}
  tACTIVITY = {}
  tPLAYER = {}
  tSAVES = {}
  tWEATHER = {}
  tHAZARDS = {}
  tSERVICES = {}
  
  initialise()
  readsettings()
  readhazards()
  readservices()
  loadmap(gMAPNAME)
end

function copyfile(filetocopy, newlocation)
  local oldfile, data, newfile
  oldfile = love.filesystem.newFile(filetocopy, "r")
  if oldfile ~= nil then
    data = oldfile:read()
    oldfile:close()
    newfile = love.filesystem.newFile(newlocation, "w")  
    newfile:write(data)
    newfile:close()
  end
end

function initialise()
  if not love.filesystem.getInfo("maps", "directory") then love.filesystem.createDirectory("maps") end
  if not love.filesystem.getInfo("tiledata", "directory") then love.filesystem.createDirectory("tiledata") end
  if not love.filesystem.getInfo("settings.txt", "file") then copyfile("data/settings.txt", "settings.txt") end
  if not love.filesystem.getInfo("hazards.txt", "file") then copyfile("data/hazards.txt", "hazards.txt") end
  if not love.filesystem.getInfo("services.txt", "file") then copyfile("data/services.txt", "services.txt") end
  if not love.filesystem.getInfo("maps/default.png", "file") then copyfile("data/default.png", "maps/default.png") end
  if not love.filesystem.getInfo("tiledata/default.txt", "file") then copyfile("data/default.txt", "tiledata/default.txt") end
end

function readsettings()
  local file = "settings.txt"
  local i = 0
  if not love.filesystem.getInfo(file, "file") then file = "data/default.txt" end
  tPLAYER = {}
  for line in love.filesystem.lines(file) do
    if string.find(line, "Police=") then gPOLICE = string.sub(line, 8)
    elseif string.find(line, "Name=") then
      i = i + 1
      tPLAYER[i] = {}
      tPLAYER[i].Name = string.sub(line, 6)
      if tPLAYER[i].Name == "Police" then tPLAYER[i].Name = "Player "..i end
      tPLAYER[i].Control = "Player"
      tPLAYER[i].R = 255
      tPLAYER[i].G = 255
      tPLAYER[i].B = 255
    elseif string.find(line, "Control=") then tPLAYER[i].Control = string.sub(line, 9)
    elseif string.find(line, "Red=") then tPLAYER[i].R = tonumber(string.sub(line, 5))
    elseif string.find(line, "Grn=") then tPLAYER[i].G = tonumber(string.sub(line, 5))
    elseif string.find(line, "Blu=") then tPLAYER[i].B = tonumber(string.sub(line, 5))
    else end
  end
  if next(tPLAYER)~=nil and gPOLICE ~= "Off" then
    tPLAYER[#tPLAYER+1] = {Name="Police", Control=gPOLICE, R=0, G=0, B=0}
  end
end

function readhazards()
  local file = "hazards.txt"
  local i = 0
  if not love.filesystem.getInfo(file, "file") then file = "data/hazards.txt" end
  tHAZARDS = {}
  for line in love.filesystem.lines(file) do
    if  string.find(line, "Name=") then
      i = i + 1
      tHAZARDS[i] = {}
      tHAZARDS[i].Name = string.sub(line, 6)
      tHAZARDS[i].Tag = ""
    elseif string.find(line, "Tag=") then tHAZARDS[i].Tag = string.sub(line, 5)
    else end
  end
end

function readservices()
  local file = "services.txt"
  local i = 0
  if not love.filesystem.getInfo(file, "file") then file = "data/services.txt" end
  tSERVICES = {}
  for line in love.filesystem.lines(file) do
    if  string.find(line, "Name=") then
      i = i + 1
      tSERVICES[i] = {}
      tSERVICES[i].Name = string.sub(line, 6)
      tSERVICES[i].Tag = ""
    elseif string.find(line, "Tag=") then tSERVICES[i].Tag = string.sub(line, 5)
    else end
  end
end

function savesettings()
  local i, j, file
  if love.filesystem.getInfo("settings.txt", "file") then love.filesystem.remove("settings.txt") end
  file = love.filesystem.newFile("settings.txt")
  file:open("w")
  file:write("Police="..gPOLICE.."\r\n")
  file:write("\r\n")
  if next(tPLAYER)~=nil then
    for i = 1, #tPLAYER, 1 do
      if tPLAYER[i].Name ~= "Police" then
        file:write("Name="..tPLAYER[i].Name.."\r\n")
        file:write("Control="..tPLAYER[i].Control.."\r\n")
        file:write("Red="..tPLAYER[i].R.."\r\n")
        file:write("Grn="..tPLAYER[i].G.."\r\n")
        file:write("Blu="..tPLAYER[i].B.."\r\n")
        file:write("\r\n")
      end
    end
  end
  file:close()
end

function loadmap(map)
  local file = "maps/"..map..".png"
  if love.filesystem.getInfo(file, "file") then
    iMAP = love.graphics.newImage(file)
    gMAPWIDTH, gMAPHEIGHT = iMAP:getWidth(), iMAP:getHeight()
    gZOOM = math.min(600/gMAPWIDTH, 600/gMAPHEIGHT)
    gZOOM_MIN = gZOOM
    gZOOM_MAX = gZOOM_MIN * 40
    gMAPNAME = map
    loadtiles(map)
    calculatelinks()
  end
end

function loadtiles(map)
  local file = "tiledata/"..map..".txt"
  local i, j, w = 0, 0, 0
  if love.filesystem.getInfo(file, "file") then
    tTILES = {}
    tACTIVITY = {}
    tWEATHER = {}
    for line in love.filesystem.lines(file) do
      if string.find(line, "Activity=") then
        table.insert(tACTIVITY, string.sub(line, 10))
      elseif string.find(line, "Weather=") then
        w = w + 1
        tWEATHER[w] = {}
        tWEATHER[w].Name = string.sub(line, 9)
        tWEATHER[w].Image = iIMAGE[w+22]
        tWEATHER[w].Tag = ""
        tWEATHER[w].Success = {}
        if next(tACTIVITY)~=nil then
          for j = 1, #tACTIVITY, 1 do
            tWEATHER[w].Success[j] = 0
          end
        end
      elseif string.find(line, "Forecast=") then tWEATHER[w].Tag = string.sub(line, 10)
      elseif string.find(line, "Type=") then
        i = i + 1
        tTILES[i] = {}
        tTILES[i].Type = tonumber(string.sub(line, 6))
        tTILES[i].Name = ""
        tTILES[i].Tag = ""
        tTILES[i].Image = iIMAGE[11+tonumber(string.sub(line, 6))]
        tTILES[i].Links = {}
        tTILES[i].Activity = {}
        if next(tACTIVITY)~=nil then
          for j = 1, #tACTIVITY, 1 do
            tTILES[i].Activity[j] = 0
          end
        end
      elseif string.find(line, "Name=") then tTILES[i].Name = string.sub(line, 6)
      elseif string.find(line, "Tag=") then tTILES[i].Tag = string.sub(line, 5)
      elseif string.find(line, "X=") then tTILES[i].X = tonumber(string.sub(line, 3))
      elseif string.find(line, "Y=") then tTILES[i].Y = tonumber(string.sub(line, 3))
      else end
      if next(tACTIVITY)~=nil then
        for j = 1, #tACTIVITY, 1 do
          if string.find(line, "Activity"..j) then tTILES[i].Activity[j] = 1
          elseif string.find(line, "Success"..j) then tWEATHER[w].Success[j] = 1
          else end
        end
      end
    end
  end
end

function savetiles(map)
  local i, j, file
  if love.filesystem.getInfo("tiledata/"..map..".txt", "file") then love.filesystem.remove("tiledata/"..map..".txt") end
  file = love.filesystem.newFile("tiledata/"..map..".txt")
  file:open("w")
  if next(tACTIVITY)~=nil then
    for i = 1, #tACTIVITY, 1 do
      file:write("Activity="..tACTIVITY[i].."\r\n")
    end
    file:write("\r\n")
  end
  if next(tWEATHER)~=nil then
    for i = 1, #tWEATHER, 1 do
      file:write("Weather="..tWEATHER[i].Name.."\r\n")
      file:write("Forecast="..tWEATHER[i].Tag.."\r\n")
      if next(tWEATHER[i].Success)~=nil then
        for j = 1, #tWEATHER[i].Success, 1 do
          if tWEATHER[i].Success[j]==1 then file:write("Success"..j.."\r\n") end
        end
      end
      file:write("\r\n")
    end
  end
  if next(tTILES)~=nil then
    for i = 1, #tTILES, 1 do
      file:write("Type="..tTILES[i].Type.."\r\n")
      if tTILES[i].Name~="" then file:write("Name="..tTILES[i].Name.."\r\n") end
      if tTILES[i].Tag~="" then file:write("Tag="..tTILES[i].Tag.."\r\n") end
      if next(tTILES[i].Activity)~= nil then
        for j = 1, #tTILES[i].Activity, 1 do
          if tTILES[i].Activity[j]~=0 then file:write("Activity"..j.."\r\n") end
        end
      end
      file:write("X="..tTILES[i].X.."\r\n")
      file:write("Y="..tTILES[i].Y.."\r\n")
      file:write("\r\n")
    end
  end
  file:close()
end

function centrecamera()
  gCAMERAX, gCAMERAY = 0, 0
  gZOOM = math.min(600/gMAPWIDTH, 600/gMAPHEIGHT)
end

function focuscamera()
  local i, j
  if next(tTILES)~=nil and next(tPLAYER)~=nil then
    i = tPLAYER[gTURN].Tile
    if i ~= 0 then
      gCAMERAX = tTILES[i].X
      gCAMERAY = tTILES[i].Y
      gZOOM = 1
    end
  end
end

function blankgame()
  local i
  if next(tPLAYER)~=nil then
    for i = 1, #tPLAYER, 1 do
      tPLAYER[i].Activities = {}
      tPLAYER[i].Convictions = {}
      tPLAYER[i].SpareTyre = 0
      tPLAYER[i].FogLights = 0
      tPLAYER[i].SnowChains = 0
      tPLAYER[i].MOT = 0
      tPLAYER[i].Insurance = 0
      tPLAYER[i].Tile = 0
      tPLAYER[i].Previous = 0
      tPLAYER[i].Moves = 0
      tPLAYER[i].LastTown = 0
      tPLAYER[i].MissTurn = 0
    end
  end
  gTURN = 0
  gTYPING = false
  gTYPEDTEXT = ""
end

function newgame()
  local cities, cards
  local i, j, k
  gPLAYERS = 0
  if next(tPLAYER)~= nil then
    cities = getplaces("City")
    for i = 1, #tPLAYER, 1 do
      if next(tACTIVITY)~=nil and tPLAYER[i].Name ~= "Police" then
        cards = {}
        for j = 1, #tACTIVITY, 1 do
          table.insert(cards, j)
        end
        for k = 1, 5, 1 do
          j = math.random(#cards)
          tPLAYER[i].Activities[k] = {}
          tPLAYER[i].Activities[k].ID = cards[j]
          tPLAYER[i].Activities[k].Complete = false
          table.remove(cards, j)
        end
      end
      if next(cities)~=nil then
        j = math.random(#cities)
        tPLAYER[i].Tile = cities[j]
        tPLAYER[i].LastTown = cities[j]
        table.remove(cities, j)
      end
      if tPLAYER[i].Control ~= "Off" and tPLAYER[i].Name ~= "Police" then gPLAYERS = gPLAYERS + 1 end
    end
    i = math.random(gPLAYERS)
    for j = 1, #tPLAYER, 1 do
      if tPLAYER[j].Control ~= "Off" then i = i - 1 end
      if i == 0 then gTURN = j end
    end
    gWEATHER = math.random(6)
    gNEXT_WEATHER = math.random(6)
    newturn()
  else 
    gGAMEMODE = "Main Menu"
    gFADE = 255
  end
end

function newturn()
  repeat
    gTURN = gTURN + 1
    if gTURN > #tPLAYER then gTURN = 1 end
    if tPLAYER[gTURN].Control ~= "Off" and tPLAYER[gTURN].Name ~= "Police" then tPLAYER[gTURN].MissTurn = tPLAYER[gTURN].MissTurn-1 end
  until tPLAYER[gTURN].MissTurn < 0 or tPLAYER[gTURN].Name == "Police"
  if tPLAYER[gTURN].Name == "Police" then playsound(6) end
  tPLAYER[gTURN].MissTurn = 0
  gDICE, gSTAGE, gLOCATION = 0, 0, 0
  gSELECT, gSUBSELECT = 0, 0
  gATTEMPTED = false
  if tPLAYER[gTURN].Control == "Player" then gGAMEMODE = "New Turn"
  else 
    gGAMEMODE = "Game"
    gDELAY = gFPS
  end
  focuscamera()
end

function love.update()
  local x, y = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  if gFADE > 0 then  -- fade in graphics
    gFADE = gFADE - gFADESPEED
    if gFADE < 0 then gFADE = 0 end
  end
  gFLASH = gFLASH + gFLASH_z  -- flash counter
  if gFLASH > 255 then  -- this affects everything that flashes (same speed)
    gFLASH = gFLASH_max
    gFLASH_z = 0 - gFLASH_z
  elseif gFLASH < gFLASH_min then
    gFLASH = gFLASH_min
    gFLASH_z = 0 - gFLASH_z
  else end
  if gGAMEMODE == "Game" and gDELAY > 0 then 
    gDELAY = gDELAY - 1
    if gSTAGE == 0 and tPLAYER[gTURN].Control == "A.I." and gDELAY == 0 then 
      gSTAGE = 1
      gDELAY = gFPS
      gDICE = math.random(6)
    elseif gSTAGE == 1 then
      if gDELAY % 4 == 0 then 
        gDICE = math.random(6)
        if not sSOUND[1]:isPlaying() then playsound(1) end
      end
      if gDELAY == 0 then
        playsound(2)
        if tPLAYER[gTURN].Name ~= "Police" then tPLAYER[gTURN].Moves = gDICE
        else tPLAYER[gTURN].Moves = gDICE * 2 end
        gSTAGE = 2
        if tPLAYER[gTURN].Control == "A.I." then gDELAY = gFPS end
        if (tTILES[tPLAYER[gTURN].Tile].Type == 2 or tTILES[tPLAYER[gTURN].Tile].Type == 6) and gDICE > 3 and tPLAYER[gTURN].Name ~= "Police" then
          table.insert(tPLAYER[gTURN].Convictions, "SP30: "..gDICE.."0mph in a 30mph limit near "..tTILES[getplaces("All", true)].Name)
        end
        if tWEATHER[gWEATHER].Name=="Foggy" and tPLAYER[gTURN].FogLights == 0 and gDICE>5 and tPLAYER[gTURN].Name ~= "Police" then
          tPLAYER[gTURN].Moves = 0
          moveplayer(gTURN, tPLAYER[gTURN].Tile, getplaces("City", true), true)
          gATTEMPTED = true
          playsound(5)
        elseif tWEATHER[gWEATHER].Name == "Snow" and tPLAYER[gTURN].SnowChains == 0 and gDICE>4 and tPLAYER[gTURN].Name ~= "Police" then
          tPLAYER[gTURN].Moves = 0
          moveplayer(gTURN, tPLAYER[gTURN].Tile, getplaces("City", true), true)
          gATTEMPTED = true
          playsound(5)
        else end
      end
    elseif gSTAGE == 2 and gDELAY == 0 and tPLAYER[gTURN].Control == "A.I." then 
      if tPLAYER[gTURN].Name == "Police" then police()
      else ai() end
    elseif gSTAGE == 3 then
      if gDELAY == 0 then
        if tTILES[tPLAYER[gTURN].Tile].Type == 3 then resolveservice(gTURN, gSERVICE)
        elseif tTILES[tPLAYER[gTURN].Tile].Type == 4 then resolvehazard(gTURN, gHAZARD)
        else end
      end
    elseif gSTAGE == 4 then
      if gDELAY % 6 == 0 then 
        gWEATHER = math.random(6) 
        if not sSOUND[1]:isPlaying() then playsound(1) end
      end
      if gDELAY == 0 then
        playsound(2)
        gWEATHER = gNEXT_WEATHER
        gNEXT_WEATHER = math.random(6)
        resolveactivity(gSUBSELECT)
      end
    else end
  end
  if gGAMEMODE == "Game" or gGAMEMODE == "Editor" then
    if love.mouse.isDown(1) then heldmouse(x,y)
    else gMOUSEHELD = 0 end
  end
  if love.timer.getTime()-gTIMER<=(1/gFPS) then love.timer.sleep((1/gFPS)-(love.timer.getTime()-gTIMER)) end  -- frame rate
  gTIMER = love.timer.getTime()  -- new timer
end

function love.draw()
  love.graphics.scale(gRATIOX, gRATIOY)
  if gGAMEMODE == "Title" then drawtitlescreen()
  elseif gGAMEMODE == "Main Menu" then drawmainmenu()
  elseif gGAMEMODE == "New Turn" then drawnewturn()
  elseif gGAMEMODE == "Game" then drawgame()
  elseif gGAMEMODE == "Editor" then draweditor()
  elseif gGAMEMODE == "Load" then drawload()
  elseif gGAMEMODE == "Edit Town" then drawtown()
  elseif gGAMEMODE == "Record" then drawconvictions()
  elseif gGAMEMODE == "Settings" then drawsettings()
  else end
  if gTYPING == true then drawtypedtext() end
  if gFADE > 0 then 
    love.graphics.setColor(0, 0, 0, gFADE)
    love.graphics.rectangle("fill", 0, 0, 800, 600)
  else end
end

function drawtitlescreen()
  love.graphics.setColor(0, 181, 255, 255)
  love.graphics.rectangle("fill", 0, 0, 800, 600)
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.draw(iIMAGE[7], 200, 250)
end

function drawmainmenu()
  local x, y = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local menu = {"play a game", "map editor", "settings", "quit"}
  local i
  love.graphics.setColor(0, 181, 255, 255)
  love.graphics.rectangle("fill", 0, 0, 800, 600)
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.draw(iIMAGE[7], 200, 40)
  for i = 1, #menu, 1 do
    if x > 300 and x < 500 and y > 125+(i*75) and y < 165+(i*75) then
      drawshadowtext(menu[i], 300, 125+(i*75), 200, "center", fTITLE, 239, 185, 11)
    else drawshadowtext(menu[i], 300, 125+(i*75), 200) end
  end
  if gMUSICON == true then drawshadowtext("Music: www.bensound.com", 10, 570, 390, "left", fSMALL) end
end

function drawnewturn()
  local r, g, b = tPLAYER[gTURN].R, tPLAYER[gTURN].G, tPLAYER[gTURN].B
  love.graphics.setColor(r, g, b, 255)
  love.graphics.rectangle("fill", 0, 0, 800, 600)
  drawshadowtext(tPLAYER[gTURN].Name.."'s turn", 200, 250, 400)
end

function drawshadowtext(text, x, y, limit, align, font, r, g, b, t)
  align = align or "center"
  font = font or fTITLE
  r, g, b, t = r or 255, g or 255, b or 255, t or 255
  love.graphics.setFont(font)
  love.graphics.setColor(0, 0, 0, t)
  love.graphics.printf(text, x+2, y+2, limit, align)
  love.graphics.setColor(r, g, b, t)
  love.graphics.printf(text, x, y, limit, align)
end

function drawgame()
  local r, g, b = tPLAYER[gTURN].R, tPLAYER[gTURN].G, tPLAYER[gTURN].B
  local x, y = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local i, j, places, found
  -- Draw map, roads, tiles, players, town names & zoom
  drawmap()
  -- Draw the prevailing weather
  love.graphics.setColor(32, 52, 36, 255)
  love.graphics.rectangle("fill", 4, 4, 192, 192)
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.rectangle("line", 4, 4, 192, 192)
  love.graphics.draw(tWEATHER[gWEATHER].Image, 36, 20)
  drawshadowtext(tWEATHER[gWEATHER].Name, 4, 8, 192, "center", fDEFAULT)
  drawshadowtext(tWEATHER[gWEATHER].Tag, 4, 132, 192, "center", fSMALL)
  -- Draw criminal symbol if they have convictions
  if next(tPLAYER[gTURN].Convictions) ~= nil then
    love.graphics.setColor(0, 0, 0, 255)
    love.graphics.draw(iIMAGE[20], 2, 520)
    if x>10 and x<70 and y>530 and y<590 then love.graphics.setColor(239, 185, 11, 255)
    else love.graphics.setColor(255, 255, 255, 255) end
    love.graphics.draw(iIMAGE[20], 0, 518)
    drawshadowtext("You have convictions", 80, 550, 590, "left", fHEADING)
  end
  -- Draw speed/moves info
  if gSTAGE >= 2 and gDICE > 0 then
    drawshadowtext(gDICE.."0mph", 400, 10, 190, "right", fHEADING)
    if tPLAYER[gTURN].Moves > 0 then drawshadowtext(tPLAYER[gTURN].Moves.." moves", 400, 60, 190, "right", fHEADING) end
  end
  -- Draw sidepanel, player name, dice
  love.graphics.setColor(r, g, b, 255)
  love.graphics.rectangle("fill", 600, 0, 200, 600)
  drawshadowtext(tPLAYER[gTURN].Name, 600, 4, 200)
  love.graphics.setColor(255, 255, 255, 255)
  if gDICE > 0 and gDICE <= 6 then
    love.graphics.draw(iIMAGE[gDICE], 650, 55)
  elseif gSTAGE == 0 then
    love.graphics.rectangle("line", 650, 55, 100, 100)
    love.graphics.rectangle("line", 651, 56, 98, 98)
    if x > 650 and x < 750 and y > 55 and y < 155 then drawshadowtext("roll the dice", 670, 55, 60, "center", fHEADING, 239, 185, 11)
    else drawshadowtext("roll the dice", 670, 55, 60, "center", fHEADING) end
  else end
  -- Draw activities
  if next(tPLAYER[gTURN].Activities)~=nil and gSTAGE >= 2 then
    drawshadowtext("Activities:", 610, 170, 190, "left", fDEFAULT, 239, 185, 11)
    for i = 1, #tPLAYER[gTURN].Activities, 1 do
      j = tPLAYER[gTURN].Activities[i].ID
      if x > 600 and x < 760 and y > (i*30)+170 and y < (i*30)+190 and tPLAYER[gTURN].Control == "Player" then
        drawshadowtext(tACTIVITY[j], 600, (i*30)+170, 160, "right", fDEFAULT, 239, 185, 11)
      else drawshadowtext(tACTIVITY[j], 600, (i*30)+170, 160, "right", fDEFAULT) end
      love.graphics.setColor(255, 255, 255, 255)
      love.graphics.circle("fill", 780, (i*30)+180, 10)
      love.graphics.setColor(0, 0, 0, 255)
      love.graphics.circle("line", 780, (i*30)+180, 10)
      if tPLAYER[gTURN].Activities[i].Complete == true then drawtickmark(780, (i*30)+180, 20)
      else drawcrossmark(780, (i*30)+180, 20) end
    end
  end
  if tPLAYER[gTURN].Name ~= "Police" then
    -- Draw lifelines (MOT/insurance) etc.
    drawshadowtext("Lifelines:", 610, 360, 190, "left", fDEFAULT, 239, 185, 11)
    drawshadowtext("Spare tyre", 600, 390, 160, "right", fDEFAULT)
    drawshadowtext("Fog lights", 600, 420, 160, "right", fDEFAULT)
    drawshadowtext("Snow chains", 600, 450, 160, "right", fDEFAULT)
    drawshadowtext("MOT certificate", 600, 480, 160, "right", fDEFAULT)
    drawshadowtext("Insurance", 600, 510, 160, "right", fDEFAULT)
    if tPLAYER[gTURN].SpareTyre > 0 then drawtickmark(780, 400, 20)
    else drawcrossmark(780, 400, 20) end
    if tPLAYER[gTURN].FogLights > 0 then drawtickmark(780, 430, 20)
    else drawcrossmark(780, 430, 20) end
    if tPLAYER[gTURN].SnowChains > 0 then drawtickmark(780, 460, 20)
    else drawcrossmark(780, 460, 20) end
    if tPLAYER[gTURN].MOT > 0 then drawtickmark(780, 490, 20)
    else drawcrossmark(780, 490, 20) end
    if tPLAYER[gTURN].Insurance > 0 then drawtickmark(780, 520, 20)
    else drawcrossmark(780, 520, 20) end
  else
    -- Draw most wanted list
    drawshadowtext("Criminal Offences:", 610, 170, 190, "left", fDEFAULT, 239, 185, 11)
    for i = 1, #tPLAYER, 1 do
      if next(tPLAYER[i].Convictions)~=nil then
        drawshadowtext(tPLAYER[i].Name, 610, (i*30)+170, 130, "right", fDEFAULT, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B)
        drawshadowtext("x"..#tPLAYER[i].Convictions, 750, (i*30)+170, 50, "left", fDEFAULT)
      end
    end
  end
  -- Draw end of turn & quit game buttons
  if gSTAGE == 2 and tPLAYER[gTURN].Control == "Player" then
    if x > 620 and x < 680 and y > 545 and y < 590 then drawtickmark(650, 567, 40)
    else drawshadowtext("End turn", 620, 545, 60, "center", fDEFAULT) end
  end
  if x > 720 and x < 780 and y > 545 and y < 590 then drawcrossmark(750, 567, 40)
  else drawshadowtext("Quit game", 720, 545, 60, "center", fDEFAULT) end
  -- Draw notice of hazard/services
  if gSTAGE == 3 then
    love.graphics.setColor(0, 0, 0, 200)
    love.graphics.rectangle("fill", 75, 380, 450, 200)
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.rectangle("line", 75, 380, 450, 200)
    if tTILES[tPLAYER[gTURN].Tile].Type == 3 then
      love.graphics.draw(iIMAGE[14], 80, 385, 0, 0.5, 0.5)
      drawshadowtext(tSERVICES[gSERVICE].Name, 120, 390, 400, "left", fHEADING)
      if tSERVICES[gSERVICE].Name~="Weather forecast" then drawshadowtext(tSERVICES[gSERVICE].Tag, 80, 440, 440, "left", fDEFAULT)
      else drawshadowtext(tSERVICES[gSERVICE].Tag.." "..tWEATHER[gNEXT_WEATHER].Name..".", 80, 440, 440, "left", fDEFAULT) end
    elseif tTILES[tPLAYER[gTURN].Tile].Type == 4 then
      love.graphics.draw(iIMAGE[15], 80, 385, 0, 0.5, 0.5)
      drawshadowtext(tHAZARDS[gHAZARD].Name, 120, 390, 400, "left", fHEADING)
      if gLOCATION == 0 then drawshadowtext(tHAZARDS[gHAZARD].Tag, 80, 440, 440, "left", fDEFAULT)
      else drawshadowtext(tHAZARDS[gHAZARD].Tag.." "..tTILES[gLOCATION].Name..".", 80, 440, 440, "left", fDEFAULT) end
    else end
  end
  -- Draw selected activity
  if gSELECT == 1 and next(tACTIVITY) ~= nil then
    love.graphics.setColor(0, 0, 0, 200)
    love.graphics.rectangle("fill", 75, 380, 450, 200)
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.rectangle("line", 75, 380, 450, 200)
    drawshadowtext(tACTIVITY[tPLAYER[gTURN].Activities[gSUBSELECT].ID], 80, 390, 440, "left", fHEADING, 239, 185, 11)
    places = getlocations(tPLAYER[gTURN].Activities[gSUBSELECT].ID)
    found = 0
    j = "Locations: "
    if next(places)~=nil then
      for i = 1, #places, 1 do
        if places[i] == tPLAYER[gTURN].Tile then found = i end
        if j == "Locations: " then j = j..tTILES[places[i]].Name
        else j = j..", "..tTILES[places[i]].Name end
      end
      if j ~= "Locations: " then j = j .. "." end
    end
    drawshadowtext(j, 80, 440, 440, "left", fDEFAULT)
    if found ~= 0 and tPLAYER[gTURN].Activities[gSUBSELECT].Complete ~= true then
      if gSTAGE == 4 then drawshadowtext("Determining weather...", 80, 540, 440, "left", fDEFAULT, 239, 185, 11)
      elseif gATTEMPTED == true then drawshadowtext("Wait until next turn", 80, 540, 440, "left", fDEFAULT)
      elseif x > 80 and x < 520 and y > 540 and y < 580 and tPLAYER[gTURN].Control == "Player" then
        drawshadowtext("Click to do: "..tACTIVITY[tPLAYER[gTURN].Activities[gSUBSELECT].ID], 80, 540, 440, "left", fDEFAULT, 239, 185, 11)
      else drawshadowtext("Attempt '"..tACTIVITY[tPLAYER[gTURN].Activities[gSUBSELECT].ID].."'?", 80, 540, 440, "left", fDEFAULT) end
    end
  end
  -- Draw end of game message
  if gSTAGE == 5 then
    love.graphics.setColor(0, 0, 0, 200)
    love.graphics.rectangle("fill", 100, 275, 400, 50)
    love.graphics.setColor(255, 255, 255, gFLASH)
    love.graphics.rectangle("line", 100, 275, 400, 50)
    drawshadowtext(tPLAYER[gTURN].Name.." has won the game", 100, 280, 400, "center", fHEADING, tPLAYER[gTURN].R, tPLAYER[gTURN].G, tPLAYER[gTURN].B)
  end
end

function drawtickmark(x, y, size)
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.rectangle("fill", x-(size/2), y-(size/2), size, size)
  --love.graphics.circle("fill", x, y, size/2)
  love.graphics.setColor(0, 0, 0, 255)
  love.graphics.rectangle("line", x-(size/2), y-(size/2), size, size)
  --love.graphics.circle("line", x, y, size/2)
  love.graphics.setColor(0, 255, 0, 255)
  love.graphics.draw(iIMAGE[22], x, y, 0, size/iIMAGE[22]:getWidth(), size/iIMAGE[22]:getHeight(), iIMAGE[22]:getWidth()/2, iIMAGE[22]:getHeight()/2)
end

function drawcrossmark(x, y, size)
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.rectangle("fill", x-(size/2), y-(size/2), size, size)
  --love.graphics.circle("fill", x, y, size/2)
  love.graphics.setColor(0, 0, 0, 255)
  love.graphics.rectangle("line", x-(size/2), y-(size/2), size, size)
  --love.graphics.circle("line", x, y, size/2)
  love.graphics.setColor(255, 0, 0, 255)
  love.graphics.draw(iIMAGE[21], x, y, 0, size/iIMAGE[21]:getWidth(), size/iIMAGE[21]:getHeight(), iIMAGE[21]:getWidth()/2, iIMAGE[21]:getHeight()/2)
end

function drawmap()
  local x, y = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local cx, cy, cz
  local i, j, k
  --Draw sea
  love.graphics.setColor(0, 181, 255, 255)
  love.graphics.rectangle("fill", 0, 0, 600, 600)
  --Draw map
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.draw(iMAP, 300-(gCAMERAX*gZOOM), 300-(gCAMERAY*gZOOM), 0, gZOOM, gZOOM, gMAPWIDTH/2, gMAPHEIGHT/2)
  if next(tTILES)~=nil and next(tPLAYER)~=nil then
    --Draw roads
    love.graphics.setLineWidth(5*gZOOM)
    for i = 1, #tTILES, 1 do
      cx, cy = 300+((tTILES[i].X-gCAMERAX)*gZOOM), 300+((tTILES[i].Y-gCAMERAY)*gZOOM)
      if next(tTILES[i].Links)~=nil then
        for j = 1, #tTILES[i].Links, 1 do
          k = tTILES[i].Links[j]
          love.graphics.line(cx, cy, 300+((tTILES[k].X-gCAMERAX)*gZOOM), 300+((tTILES[k].Y-gCAMERAY)*gZOOM))
        end
      end
    end
    love.graphics.setLineWidth(1)
    --Draw tiles
    for i = 1, #tTILES, 1 do
      cx, cy = 300+((tTILES[i].X-gCAMERAX)*gZOOM), 300+((tTILES[i].Y-gCAMERAY)*gZOOM)
      love.graphics.setColor(255, 255, 255, 255)
      -- Flashes if the current player is on this tile
      if gTURN ~= 0 then
        if tPLAYER[gTURN].Tile == i then love.graphics.setColor(255, 255, 255, gFLASH) end
      end
      -- Also flashes if the mouse is on the tile and it's available as a move
      if gTURN ~= 0 and gSTAGE == 2 then
        if getisconnected(tPLAYER[gTURN].Tile, i) and distancebetween(x, y, cx, cy) < tTILES[i].Image:getWidth()*gZOOM/2 and i ~= tPLAYER[gTURN].Previous and tPLAYER[gTURN].Moves > 0 and tPLAYER[gTURN].Control == "Player" then
          love.graphics.setColor(255, 255, 255, gFLASH)
        end
      end
      love.graphics.draw(tTILES[i].Image, cx, cy, 0, gZOOM, gZOOM, tTILES[i].Image:getWidth()/2, tTILES[i].Image:getHeight()/2)
    end
    --Draw player icons
    for i = 1, #tPLAYER, 1 do
      cz = tPLAYER[i].Tile
      if cz ~= 0 and i ~= gTURN and tPLAYER[i].Control ~= "Off" then
        cx, cy = 300+((tTILES[cz].X-gCAMERAX)*gZOOM), 300+((tTILES[cz].Y-gCAMERAY)*gZOOM)
        love.graphics.setColor(0, 0, 0, 255)
        love.graphics.draw(iIMAGE[11], cx+2, cy+2, 0, gZOOM, gZOOM, iIMAGE[11]:getWidth()/2, iIMAGE[11]:getHeight()/2)
        love.graphics.setColor(tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B, 255)
        love.graphics.draw(iIMAGE[11], cx, cy, 0, gZOOM, gZOOM, iIMAGE[11]:getWidth()/2, iIMAGE[11]:getHeight()/2)
      end
    end
    --Draw the player whose turn it is last
    if gTURN ~= 0 then
      cz = tPLAYER[gTURN].Tile
      if cz ~= 0 then
        cx, cy = 300+((tTILES[cz].X-gCAMERAX)*gZOOM), 300+((tTILES[cz].Y-gCAMERAY)*gZOOM)
        love.graphics.setColor(0, 0, 0, 255)
        love.graphics.draw(iIMAGE[11], cx+2, cy+2, 0, gZOOM, gZOOM, iIMAGE[11]:getWidth()/2, iIMAGE[11]:getHeight()/2)
        love.graphics.setColor(tPLAYER[gTURN].R, tPLAYER[gTURN].G, tPLAYER[gTURN].B, 255)
        love.graphics.draw(iIMAGE[11], cx, cy, 0, gZOOM, gZOOM, iIMAGE[11]:getWidth()/2, iIMAGE[11]:getHeight()/2)
      end
    end
    --Draw names of cities/towns/jail on top
    love.graphics.setFont(fHEADING)
    for i = 1, #tTILES, 1 do
      cx, cy = 300+((tTILES[i].X-gCAMERAX)*gZOOM), 300+((tTILES[i].Y-gCAMERAY)*gZOOM)
      love.graphics.setColor(0, 0, 0, 255)
      love.graphics.printf(tTILES[i].Name, cx, cy, 160, "center", 0, gZOOM, gZOOM, 78, -32)
      love.graphics.setColor(255, 255, 255, 255)
      love.graphics.printf(tTILES[i].Name, cx, cy, 160, "center", 0, gZOOM, gZOOM, 80, -30)
    end
  end
  -- Draw zoom symbols
  love.graphics.setColor(0, 0, 0, 255)
  love.graphics.draw(iIMAGE[8], 551, 451)
  love.graphics.draw(iIMAGE[9], 551, 501)
  love.graphics.draw(iIMAGE[10], 551, 551)
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.draw(iIMAGE[8], 550, 450)
  love.graphics.draw(iIMAGE[9], 550, 500)
  love.graphics.draw(iIMAGE[10], 550, 550)
end

function draweditor()
  local x, y = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local cx, cy = gCAMERAX+((x - 300)/gZOOM), gCAMERAY+((y - 300)/gZOOM)
  drawmap()
  drawpanel()
  drawbrush()
end

function drawpanel()
  local x, y = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local i
  love.graphics.setColor(40, 40, 40, 255)
  love.graphics.rectangle("fill", 600, 0, 200, 600)
  drawshadowtext("tools", 600, 20, 200)
  if x > 650 and x < 750 and y > 540 and y < 580 then drawshadowtext("exit", 650, 540, 100, "center", fTITLE, 239, 185, 11)
  else drawshadowtext("exit", 650, 540, 100) end
  if x > 605 and x < 695 and y > 465 and y < 505 then drawshadowtext("load", 600, 465, 100, "center", fTITLE, 239, 185, 11)
  else drawshadowtext("load", 600, 465, 100) end
  if x > 705 and x < 795 and y > 465 and y < 505 then drawshadowtext("save", 700, 465, 100, "center", fTITLE, 239, 185, 11)
  else drawshadowtext("save", 700, 465, 100) end
  for i = 1, 8, 1 do
    if gBRUSH == i then love.graphics.setColor(255, 255, 255, gFLASH)
    else love.graphics.setColor(255, 255, 255, 255) end
    love.graphics.draw(iIMAGE[11+i], 618+(math.floor(i/5)*100), (i*90)-(math.floor(i/5)*360))
  end
end

function drawbrush()
  local x, y = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  if x > 0 and x < 600 and y > 0 and y < 600 and gBRUSH ~= 0 and gBRUSH ~= 8 then
    love.graphics.setColor(255, 255, 255, 100)
    love.graphics.draw(iIMAGE[11+gBRUSH], x, y, 0, gZOOM, gZOOM, iIMAGE[11+gBRUSH]:getWidth()/2, iIMAGE[11+gBRUSH]:getHeight()/2)
  end
end

function drawtypedtext()
  love.graphics.setColor(0, 0, 0, 200)
  love.graphics.rectangle("fill", 120, 270, 360, 60)
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.rectangle("line", 120, 270, 360, 60)
  drawshadowtext(gTYPEDTEXT, 130, 280, 340)
end

function drawload()
  local x, y = love.mouse.getX() / gRATIOX, love.mouse.getY() / gRATIOY
  local i = 0
  local n
  if next(tSAVES)~=nil then n=#tSAVES else n=0 end
  drawshadowtext("Load", 300, 10, 200)
  love.graphics.setColor(255, 255, 255, 255)
  love.graphics.rectangle("line", 150, 60, 500, math.min(n,10)*50)
  love.graphics.setFont(fHEADING)
  if next(tSAVES)~=nil then
    repeat
      i = i + 1
      if x>160 and x<390 and y>((i-1)*50)+70 and y<(i*50)+50 and gTYPING==false then love.graphics.setColor(255, 255, 255, gFLASH)
      else love.graphics.setColor(255, 215, 0, 255) end
      love.graphics.printf(tSAVES[i], 160, ((i-1)*50)+70, 230, "left")
    until i==#tSAVES or i==10
    if i<#tSAVES then
      repeat
        i = i + 1
        if x>410 and x<640 and y>((i-11)*50)+70 and y<((i-10)*50)+50 and gTYPING==false then love.graphics.setColor(255, 255, 255, gFLASH)
        else love.graphics.setColor(255, 215, 0, 255) end
        love.graphics.printf(tSAVES[i], 410, ((i-11)*50)+70, 230, "left")
      until i==#tSAVES or i==19
    end
  end
end

function drawtown()
  local x, y = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local i
  if next(tTILES)~=nil and gSELECT ~= 0 then
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.draw(iIMAGE[11+tTILES[gSELECT].Type], 8, 8)
    drawshadowtext(tTILES[gSELECT].Name, 80, 25, 712, "left")
    drawshadowtext(tTILES[gSELECT].Tag, 80, 80, 712, "left", fHEADING)
    if next(tTILES[gSELECT].Activity)~=nil then
      for i = 1, #tACTIVITY, 1 do
        if tTILES[gSELECT].Activity[i]~=nil then
          if tTILES[gSELECT].Activity[i] == 0 then
            love.graphics.setColor(255, 0, 0, 255)
            love.graphics.draw(iIMAGE[21], 405, (i*50)+80)
          else
            love.graphics.setColor(0, 255, 0, 255)
            love.graphics.draw(iIMAGE[22], 405, (i*50)+80)
          end
        end
      end
    end    
  end
  if next(tACTIVITY)~=nil then
    for i = 1, #tACTIVITY, 1 do
      drawshadowtext(tACTIVITY[i], 8, (i*50)+80, 392, "right")
    end
  end
  if x > 700 and x < 800 and y > 550 and y < 600 then drawshadowtext("Done", 700, 550, 100, "center",  fTITLE, 239, 185, 11)
  else drawshadowtext("Done", 700, 550, 100) end
end

function drawconvictions()
  local i
  drawshadowtext(tPLAYER[gSELECT].Name..": Criminal Convictions", 10, 10, 790, "left", fTITLE, 293, 185, 11)
  if next(tPLAYER[gSELECT].Convictions)~=nil then
    for i = 1, #tPLAYER[gSELECT].Convictions, 1 do
      drawshadowtext(tPLAYER[gSELECT].Convictions[i], 10, (i*30)+50, 790, "left", fDEFAULT)
    end
  end
end

function drawsettings()
  local x, y = love.mouse.getX()/gRATIOX, love.mouse.getY()/gRATIOY
  local i
  drawshadowtext("Settings", 10, 10, 390, "left", fTITLE, 293, 185, 11)
  if x>560 and x<790 and y>10 and y<50 then drawshadowtext("Police="..gPOLICE, 400, 10, 390, "right", fTITLE, 293, 185, 11)
  else drawshadowtext("Police="..gPOLICE, 400, 10, 390, "right") end
  if x>10 and x<260 and y>550 and y<590 then drawshadowtext("Back to Main Menu", 10, 550, 250, "center", fHEADING, 239, 185, 11)
  else drawshadowtext("Back to Main Menu", 10, 550, 250, "center", fHEADING) end
  if next(tPLAYER)~=nil then
    for i = 1, #tPLAYER, 1 do
      if tPLAYER[i].Name ~= "Police" then
        if tPLAYER[i].Control ~= "Off" then 
          if x>10 and x<290 and y>(i*50)+50 and y<(i*50)+90 then
            drawshadowtext(tPLAYER[i].Name, 10, (i*50)+50, 280, "left", fHEADING, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B, gFLASH)
          else drawshadowtext(tPLAYER[i].Name, 10, (i*50)+50, 280, "left", fHEADING, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B) end
          if x>300 and x<400 and y>(i*50)+50 and y<(i*50)+90 then
            drawshadowtext(tPLAYER[i].Control, 300, (i*50)+50, 100, "center", fHEADING, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B, gFLASH)
          else drawshadowtext(tPLAYER[i].Control, 300, (i*50)+50, 100, "center", fHEADING, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B) end
          if x>500 and x<600 and y>(i*50)+50 and y<(i*50)+90 then
            drawshadowtext("R: "..tPLAYER[i].R, 500, (i*50)+54, 100, "left", fDEFAULT, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B, gFLASH)
          else drawshadowtext("R: "..tPLAYER[i].R, 500, (i*50)+54, 100, "left", fDEFAULT, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B) end
          if x>600 and x<700 and y>(i*50)+50 and y<(i*50)+90 then 
            drawshadowtext("G: "..tPLAYER[i].G, 600, (i*50)+54, 100, "left", fDEFAULT, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B, gFLASH)
          else drawshadowtext("G: "..tPLAYER[i].G, 600, (i*50)+54, 100, "left", fDEFAULT, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B) end
          if x>700 and x<800 and y>(i*50)+50 and y<(i*50)+90 then 
            drawshadowtext("B: "..tPLAYER[i].B, 700, (i*50)+54, 100, "left", fDEFAULT, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B, gFLASH)
          else drawshadowtext("B: "..tPLAYER[i].B, 700, (i*50)+54, 100, "left", fDEFAULT, tPLAYER[i].R, tPLAYER[i].G, tPLAYER[i].B) end
        else drawshadowtext("Click to add player", 10, (i*50)+50, 280, "left", fHEADING) end
      end
    end
  end
end

function love.mousepressed(x, y, button, istouch)
  x = x / gRATIOX
  y = y / gRATIOY
  if gGAMEMODE == "Title" then
    gGAMEMODE = "Main Menu"
    gFADE = 255
    gMUSIC:play()
  elseif gGAMEMODE == "Main Menu" then clickmainmenu(x, y)
  elseif gGAMEMODE == "Game" then clickgame(x, y)
  elseif gGAMEMODE == "Editor" then clickeditor(x, y)
  elseif gGAMEMODE == "Load" then clickload(x, y)
  elseif gGAMEMODE == "Edit Town" then clicktown(x, y)
  elseif gGAMEMODE == "Settings" then clicksettings(x, y)
  elseif gGAMEMODE == "New Turn" or gGAMEMODE == "Record" then
    gGAMEMODE = "Game"
    gSELECT = 0
    gFADE = 255
  else end
end

function heldmouse(x,y)
  gMOUSEHELD = gMOUSEHELD + 1
  if x > 550 and x < 582 and y > 450 and y < 482 then     -- zoom in
    gZOOM = gZOOM * gZOOMSPEED
    gZOOM = math.min(gZOOM, gZOOM_MAX)
  elseif x > 550 and x < 582 and y > 550 and y < 582 then -- zoom out
    gZOOM = gZOOM / gZOOMSPEED
    gZOOM = math.max(gZOOM, gZOOM_MIN)
  else end
end

function clickmainmenu(x, y)
  if x > 300 and x < 500 and y > 200 and y < 240 then
    gFADE = 255
    blankgame()
    newgame()
  elseif x > 300 and x < 500 and y > 275 and y < 315 then
    gFADE = 255
    blankgame()
    centrecamera()
    gGAMEMODE = "Editor"
    gSELECT, gSUBSELECT = 0, 0
  elseif x > 300 and x < 500 and y > 350 and y < 400 then
    gFADE = 255
    gGAMEMODE = "Settings"
    gSELECT, gSUBSELECT = 0, 0
  elseif x > 300 and x < 500 and y > 425 and y < 465 then
    love.timer.sleep(0.5)
    love.event.quit()
  end
end

function clickgame(x,y)
  local i, m, cx, cy, places, found
  if gSELECT == 1 and tPLAYER[gTURN].Control == "Player" then                     -- An activity has been selected, what are they clicking?
    if x > 75 and x < 525 and y > 380 and y < 580 then
      places = getlocations(tPLAYER[gTURN].Activities[gSUBSELECT].ID)
      found = 0
      if next(places)~=nil then
        for i = 1, #places, 1 do
          if places[i] == tPLAYER[gTURN].Tile then found = i end
        end
      end
      if x > 80 and x < 520 and y > 540 and y < 580 and tPLAYER[gTURN].Control == "Player" and found ~=0 then  -- Attempted to do an activity
        gDELAY = gFPS*2
        gSTAGE = 4
      end
    else
      gSELECT = 0
      gSUBSELECT = 0
    end
  elseif gSTAGE == 3 and tPLAYER[gTURN].Control == "Player" then              -- Click to make hazard/service message disappear
    if tTILES[tPLAYER[gTURN].Tile].Type == 3 then resolveservice(gTURN, gSERVICE)
    elseif tTILES[tPLAYER[gTURN].Tile].Type == 4 then resolvehazard(gTURN, gHAZARD)
    else end
  elseif gSTAGE==2 and x>600 and y<800 and y>200 and y<340 and tPLAYER[gTURN].Control == "Player" and tPLAYER[gTURN].Name~="Police" then  -- Select activity
    for i = 1, 5, 1 do
      if y > (i*30)+170 and y < (i*30)+190 then
        gSELECT = 1
        gSUBSELECT = i
      end
    end
  elseif x>610 and x<740 and y>200 and y<540 and tPLAYER[gTURN].Control == "Player" and tPLAYER[gTURN].Name=="Police" then  -- Check criminal record
    for i = 1, #tPLAYER-1, 1 do
      if y > (i*30)+170 and y < (i*30)+190 then
        gSELECT = i
      end
    end
    if gSELECT ~= 0 and next(tPLAYER[gSELECT].Convictions)~=nil then
      gGAMEMODE = "Record"
      gFADE = 255
    else gSELECT = 0 end
  elseif x > 10 and x < 70 and y > 530 and y < 590 and next(tPLAYER[gTURN].Convictions)~=nil then     -- Clicked criminal record
    gGAMEMODE = "Record"
    gSELECT = gTURN
    gFADE = 255
  elseif x > 620 and x < 680 and y > 545 and y < 590 and tPLAYER[gTURN].Control == "Player" and gSTAGE == 2 then  -- clicked end turn
    newturn()
  elseif x > 720 and x < 780 and y > 545 and y < 590 then   -- clicked quit game
    gFADE = 255
    gGAMEMODE = "Main Menu"
    gSELECT, gSUBSELECT = 0, 0
  elseif x > 650 and x < 750 and y > 50 and y < 150 and gSTAGE == 0 and tPLAYER[gTURN].Control == "Player" then  -- clicked dice
    gSTAGE = 1
    gDELAY = gFPS*2
    gDICE = math.random(6)
  elseif x > 550 and x < 582 and y > 450 and y < 482 then               -- zoom in
    -- zoom in, do nothing
  elseif x > 550 and x < 582 and y > 550 and y < 582 then               -- zoom out
    -- zoom out, do nothing
  elseif x > 550 and x < 582 and y > 500 and y < 532 then               -- clicked centre map
    centrecamera()
  elseif gSTAGE == 2 and tPLAYER[gTURN].Control == "Player" then        -- click to move
    if next(tTILES)~=nil and tPLAYER[gTURN].Control == "Player" then
      m = 0
      for i = 1, #tTILES, 1 do
        cx, cy = 300+((tTILES[i].X-gCAMERAX)*gZOOM), 300+((tTILES[i].Y-gCAMERAY)*gZOOM)
        if getisconnected(tPLAYER[gTURN].Tile, i) and distancebetween(x, y, cx, cy) < tTILES[i].Image:getWidth()*gZOOM/2 and i ~= tPLAYER[gTURN].Previous and tPLAYER[gTURN].Moves > 0 and tPLAYER[gTURN].Control == "Player" then m = i end
      end
      if m > 0 then moveplayer(gTURN, tPLAYER[gTURN].Tile, m, false) end
    end
  else end
end

function clickeditor(x,y)
  local cx, cy
  local i
  if x > 600 and x < 800 then
    if x > 618 and x < 682 and y > 90 and y < 170 then gBRUSH = 1 
    elseif x > 618 and x < 682 and y > 180 and y < 260 then gBRUSH = 2
    elseif x > 618 and x < 682 and y > 270 and y < 350 then gBRUSH = 3
    elseif x > 618 and x < 682 and y > 360 and y < 440 then gBRUSH = 4
    elseif x > 718 and x < 782 and y > 90 and y < 170 then gBRUSH = 5
    elseif x > 718 and x < 782 and y > 180 and y < 260 then gBRUSH = 6
    elseif x > 718 and x < 782 and y > 270 and y < 350 then gBRUSH = 7
    elseif x > 718 and x < 782 and y > 360 and y < 440 then gBRUSH = 8
    elseif x > 605 and x < 695 and y > 465 and y < 505 then 
      gGAMEMODE = "Load"
      gFADE = 255
    elseif x > 705 and x < 795 and y > 465 and y < 505 then
      savetiles(gMAPNAME)
    elseif x > 650 and x < 750 and y > 540 and y < 580 then
      gFADE = 255
      gGAMEMODE = "Main Menu"
      gSELECT, gSUBSELECT = 0, 0
    else gBRUSH = 0 end
  elseif x > 550 and x < 582 and y > 450 and y < 482 then
    -- zoom in, do nothing
  elseif x > 550 and x < 582 and y > 550 and y < 582 then
    -- zoom out, do nothing
  elseif x > 550 and x < 582 and y > 500 and y < 532 then
    centrecamera()
  elseif x > 0 and x < 600 and gBRUSH ~= 0 then
    if gBRUSH == 8 then erasetile(x, y)
    else createtile(x, y, gBRUSH) end
    calculatelinks()
  elseif x > 0 and x < 600 and gBRUSH == 0 then
    if next(tTILES)~=nil then
      for i = 1, #tTILES, 1 do
        cx, cy = gCAMERAX+((x - 300)/gZOOM), gCAMERAY+((y - 300)/gZOOM)
        if distancebetween(cx, cy, tTILES[i].X, tTILES[i].Y) < tTILES[i].Image:getWidth() / 2 and (tTILES[i].Type == 5 or tTILES[i].Type == 6) then 
          gGAMEMODE = "Edit Town"
          gSELECT = i
        end
      end
    end
  else end
end

function clickload(x,y)
  local f = 0
  local n
  if next(tSAVES)~=nil then n = #tSAVES else n = 0 end
  if x>160 and x<390 and y>70 and y<550 then
    f = math.floor((y-20)/50)
  elseif x>410 and x<540 and y>70 and y<550 then
    f = math.floor((y-20)/50)+10
  else end
  if f>0 and f<=n then 
    loadmap(tSAVES[f])
    gFADE = 255
    gGAMEMODE = "Editor"
  end
end

function clicktown(x,y)
  local i
  if x > 700 and x < 800 and y > 550 and  y  < 600 then
    gGAMEMODE = "Editor"
    gFADE  = 255
    gSELECT = 0
  elseif x > 400 and x < 440 and y > 130 and next(tTILES)~=nil and gSELECT~=0 and next(tACTIVITY)~=nil then
    i = math.floor((y - 80)/50)
    if i<=#tACTIVITY then
      if tTILES[gSELECT].Activity[i] == nil then tTILES[gSELECT].Activity[i] = 1
      elseif tTILES[gSELECT].Activity[i] == 1 then tTILES[gSELECT].Activity[i] = 0
      else tTILES[gSELECT].Activity[i] = 1 end
    end
  elseif x > 0 and x < 400 and y > 130 and next(tACTIVITY)~=nil then
    i = math.floor((y - 80)/50)
    if i<=#tACTIVITY then gSUBSELECT = i
    else gSUBSELECT = i + 1 end
    gTYPING = true
  else end
end

function clicksettings(x, y)
  local i, j = 0, 0
  if x>10 and x<260 and y>550 and y<590 then
    savesettings()
    gFADE = 255
    gGAMEMODE = "Main Menu"
    gSELECT, gSUBSELECT = 0, 0
  elseif x>560 and x<790 and y>10 and y<50 then
    if gPOLICE == "A.I." then 
      gPOLICE = "Player"
      if next(tPLAYER)~=nil then tPLAYER[#tPLAYER].Control = "Player" end
    elseif gPOLICE == "Player" then 
      gPOLICE = "Off"
      if next(tPLAYER)~=nil then table.remove(tPLAYER, #tPLAYER) end
    else 
      gPOLICE = "A.I."
      if next(tPLAYER)~=nil then 
        tPLAYER[#tPLAYER+1] = {Name="Police", Control=gPOLICE, R=0, G=0, B=0}
      end
    end
  elseif next(tPLAYER)~=nil then
    for i = 1, #tPLAYER, 1 do
      if y > (i*50)+50 and y < (i*50)+90 and tPLAYER[i].Name ~= "Police" then j = i end
    end
    if j~=0 then
      if tPLAYER[j].Control == "Off" then tPLAYER[j].Control = "Player"
      elseif x > 10 and x < 290 then
        gTYPING = true
        gTYPEDTEXT = tPLAYER[j].Name
        gSELECT = j
        gSUBSELECT = 1
      elseif x > 300 and x < 400 then
        if tPLAYER[j].Control == "Player" then tPLAYER[j].Control = "A.I."
        else tPLAYER[j].Control = "Off" end
      elseif x > 500 and x < 600 then
        gTYPING = true
        gTYPEDTEXT = ""
        gSELECT = j
        gSUBSELECT = 2
      elseif x > 600 and x < 700 then
        gTYPING = true
        gTYPEDTEXT = ""
        gSELECT = j
        gSUBSELECT = 3
      elseif x > 700 and x < 800 then
        gTYPING = true
        gTYPEDTEXT = ""
        gSELECT = j
        gSUBSELECT = 4
      else end
    end
  end
end

function love.mousereleased(x, y, button, istouch)
  x = x / gRATIOX
  y = y / gRATIOY
end

function love.mousemoved( x, y, dx, dy, istouch )
  if gMOUSEHELD > gFPS * 0.2 then
    gCAMERAX = gCAMERAX - (dx/gZOOM)
    gCAMERAY = gCAMERAY - (dy/gZOOM)
  end
end

function love.keypressed(key)
  if key == "escape" and gTYPING == true then
    gTYPING = false
  elseif gGAMEMODE == "Title" then
    gGAMEMODE = "Main Menu"
    gFADE = 255
    gMUSIC:play()
  elseif key == "escape" and gGAMEMODE ~= "Main Menu" then
    if gGAMEMODE == "Editor" or gGAMEMODE == "Game" or gGAMEMODE == "Settings" then
      gGAMEMODE = "Main Menu"
      gFADE = 255
      gSELECT, gSUBSELECT = 0, 0
    elseif gGAMEMODE == "Load" or gGAMEMODE == "Edit Town" then
      gGAMEMODE = "Editor"
      gFADE = 255
      gSELECT = 0
    elseif gGAMEMODE == "Record" then
      gGAMEMODE = "Game"
      gFADE = 255
    else end
  elseif gGAMEMODE == "New Turn" then
    gGAMEMODE = "Game"
    gFADE = 255
  elseif key == "backspace" and gTYPING == true then
    gTYPEDTEXT = string.sub(gTYPEDTEXT, 1, string.len(gTYPEDTEXT)-1)
  elseif key == "return" and gTYPING == true then
    if gGAMEMODE == "Editor" then
      tTILES[gSELECT].Name = gTYPEDTEXT
      gTYPING = false
      gGAMEMODE = "Edit Town"
      gFADE = 255
    elseif gGAMEMODE == "Edit Town" then
      if gTYPEDTEXT == "" then table.remove(tACTIVITY, gSUBSELECT)
      else tACTIVITY[gSUBSELECT] = gTYPEDTEXT end
      gTYPING = false
      gSUBSELECT = 0
    elseif gGAMEMODE == "Settings" then
      if gSUBSELECT == 1 and gTYPEDTEXT ~= "Police" then tPLAYER[gSELECT].Name = gTYPEDTEXT
      elseif gSUBSELECT == 2 then tPLAYER[gSELECT].R = math.max(math.min(tonumber(gTYPEDTEXT), 255), 0)
      elseif gSUBSELECT == 3 then tPLAYER[gSELECT].G = math.max(math.min(tonumber(gTYPEDTEXT), 255), 0)
      elseif gSUBSELECT == 4 then tPLAYER[gSELECT].B = math.max(math.min(tonumber(gTYPEDTEXT), 255), 0)
      else end
      gTYPING = false
      gSELECT, gSUBSELECT = 0, 0
    else end
  elseif key == "return" and gGAMEMODE == "Game" then
    if tPLAYER[gTURN].Control == "Player" and (gSTAGE == 0 or gSTAGE == 2) then newturn() end 
  else end
end

function love.textinput(text)
  if gTYPING==true then
    if gTYPEDTEXT == "enter text" then gTYPEDTEXT = "" end
    if gGAMEMODE == "Settings" and next(tPLAYER)~=nil then
      if gTYPEDTEXT == tPLAYER[gSELECT].Name and gSUBSELECT == 1 then gTYPEDTEXT = "" end
    end
    gTYPEDTEXT = gTYPEDTEXT..text
    if string.len(gTYPEDTEXT)>19 then gTYPEDTEXT=string.sub(gTYPEDTEXT,-19) end
  end
end

function createtile(x, y, b)
  local cx, cy = gCAMERAX+((x - 300)/gZOOM), gCAMERAY+((y - 300)/gZOOM)
  local i, j
  if next(tTILES) ~= nil then i = #tTILES + 1 else i = 1 end
  tTILES[i] = {}
  tTILES[i].Name = ""
  tTILES[i].Tag = ""
  if b == 5 or b == 6 or b == 7 then  -- Name these tiles
    gTYPEDTEXT = "enter text"
    gTYPING = true
    gSELECT = i
  end
  tTILES[i].Type = b
  tTILES[i].X = math.floor(cx+0.5)
  tTILES[i].Y = math.floor(cy+0.5)
  tTILES[i].Image = iIMAGE[b+11]
  tTILES[i].Links = {}
  tTILES[i].Activity = {}
  if next(tACTIVITY)~=nil then
    for j = 1, #tACTIVITY, 1 do
      tTILES[i].Activity[j] = 0
    end
  end
end

function erasetile(x, y)
  local cx, cy = gCAMERAX+((x - 300)/gZOOM), gCAMERAY+((y - 300)/gZOOM)
  local i
  if next(tTILES) ~= nil then
    for i = #tTILES, 1, -1 do
      if distancebetween(cx, cy, tTILES[i].X, tTILES[i].Y) <= tTILES[i].Image:getWidth()/2 then table.remove(tTILES, i) end
    end
  end
end

function calculatelinks()
  local i, j
  if next(tTILES)~=nil then
    for i = 1, #tTILES, 1 do
      tTILES[i].Links = {}
      for j = 1, #tTILES, 1 do        
        if i ~= j and distancebetween(tTILES[i].X, tTILES[i].Y, tTILES[j].X, tTILES[j].Y) < tTILES[i].Image:getWidth() * 1.3 then 
          table.insert(tTILES[i].Links, j)
        end
      end
    end
  end
end

function distancebetween(a, b, c, d)
  return math.sqrt((a-c)^2+(b-d)^2)
end

function getplayersbytile(t)
  local p = {}
  local i
  if next(tPLAYER)~=nil then
    for i = 1, #tPLAYER, 1 do
      if tPLAYER[i].Tile == t then table.insert(p, i) end
    end
  end
  return p
end

function getisconnected(a, b)
  local connected = false
  local i
  if next(tTILES)~=nil then
    if next(tTILES[a].Links) ~= nil then
      for i = 1, #tTILES[a].Links, 1 do
        if tTILES[a].Links[i] == b then connected = true end
      end
    end
    if next(tTILES[b].Links) ~= nil then
      for i = 1, #tTILES[b].Links, 1 do
        if tTILES[b].Links[i] == a then connected = true end
      end
    end
  end
  return connected
end

function getplaces(kind, nearest)
  nearest = nearest or false
  local places = {}
  local x, y, record, distance
  local i
  if next(tTILES)~=nil then
    if nearest == true then
      x, y = tTILES[tPLAYER[gTURN].Tile].X, tTILES[tPLAYER[gTURN].Tile].Y
      record = math.huge
    end
    for i = 1, #tTILES, 1 do
      if tTILES[i].Type == 5 and (kind == "All" or kind == "Town") then 
        if nearest == true then
          distance = distancebetween(x, y, tTILES[i].X, tTILES[i].Y)
          if distance < record then
            places = i
            record = distance
          end
        else table.insert(places, i) end
      elseif tTILES[i].Type == 6 and (kind == "All" or kind == "City") then 
        if nearest == true then
          distance = distancebetween(x, y, tTILES[i].X, tTILES[i].Y)
          if distance < record then
            places = i
            record = distance
          end
        else table.insert(places, i) end
      else end
    end
  end
  return places
end

function getlocations(n)
  local locs = {}
  local i
  if next(tTILES)~=nil and next(tACTIVITY)~=nil then
    for i = 1, #tTILES, 1 do
      if tTILES[i].Type == 5 or tTILES[i].Type == 6 then
        if next(tTILES[i].Activity)~=nil then
          if tTILES[i].Activity[n] == 1 then table.insert(locs, i) end
        end
      end
    end
  end
  return locs
end

function moveplayer(p, start, finish, jump)
  local places = getplaces("City")
  local i, j, jail
  if jump == true or tTILES[finish].Type == 5 or tTILES[finish].Type == 6 or tTILES[finish].Type == 7 then tPLAYER[p].Previous = finish
  else tPLAYER[p].Previous = start end
  tPLAYER[p].Tile = finish
  if jump ~= true then
    if (tTILES[finish].Type==2 or tTILES[finish].Type==6) and gDICE>3 and tPLAYER[gTURN].Name~="Police" then
      table.insert(tPLAYER[p].Convictions, "SP30: "..gDICE.."0mph in a 30mph limit near "..tTILES[getplaces("All", true)].Name)
    end
    if (tTILES[finish].Type==1 or tTILES[finish].Type==3 or tTILES[finish].Type==4 or tTILES[finish].Type==5) and gDICE> 6 and tPLAYER[gTURN].Name~="Police" then
      table.insert(tPLAYER[p].Convictions, "SP50: "..gDICE.."0mph in a 60mph limit near "..tTILES[getplaces("All", true)].Name)
    end
    tPLAYER[p].Moves = math.max(tPLAYER[p].Moves - 1, 0)
  end
  if tTILES[finish].Type == 5 or tTILES[finish].Type == 6 then tPLAYER[p].LastTown = finish end
  if (tTILES[finish].Type == 3 or tTILES[finish].Type == 4) and tPLAYER[gTURN].Name~="Police" then
    gSTAGE = 3
    if tTILES[finish].Type == 3 then 
      gSERVICE = math.random(#tSERVICES)
      if tWEATHER[gWEATHER].Name == "Blistering Heat" then tPLAYER[p].Moves = 0 end
    else 
      gHAZARD = math.random(#tHAZARDS)
      if tHAZARDS[gHAZARD].Name == "Family birthday" then
        if next(places)~=nil then gLOCATION = places[math.random(#places)] end
      end
    end
    if tPLAYER[p].Control == "A.I." then gDELAY = gFPS*3 end
  else
    if tPLAYER[p].Control == "A.I." then gDELAY = math.ceil(gFPS/2) end
  end
  -- Has anyone been caught by the police?
  if gPOLICE ~= "Off" then
    for i = 1, #tPLAYER-1, 1 do
      if tPLAYER[#tPLAYER].Tile == tPLAYER[i].Tile and next(tPLAYER[i].Convictions)~=nil then
        playsound(6)
        jail = {}
        if next(tTILES)~=nil then
          for j = 1, #tTILES, 1 do
            if tTILES[j].Name == "Jail" then table.insert(jail, j) end
          end
        end
        if next(jail)~=nil then
          tPLAYER[i].Convictions = {}
          tPLAYER[i].Tile = jail[math.random(#jail)]
          tPLAYER[i].Previous = tPLAYER[i].Tile
        end
      end
    end
  end
end

function resolvehazard(p, h)
  local i, j
  if tHAZARDS[h].Name == "Car insurance renewal" then
    if tPLAYER[p].Insurance == 0 then
      table.insert(tPLAYER[p].Convictions, "IN10: Using a vehicle uninsured against third party risks")
    else tPLAYER[p].Insurance = 0 end
    gSTAGE = 2
  elseif tHAZARDS[h].Name == "Family argument" then
    if next(tACTIVITY)~=nil and next(tPLAYER[p].Activities)~=nil then
      for i = 1, #tPLAYER[p].Activities, 1 do
        if tPLAYER[p].Activities[i].Complete == false then
          j = math.random(#tACTIVITY)
          tPLAYER[p].Activities[i].ID = j
        end
      end
    end
    gSTAGE = 2
  elseif tHAZARDS[h].Name == "Family birthday" then
    if gLOCATION ~= 0 then moveplayer(p, tPLAYER[p].Tile, gLOCATION, true) end
    newturn()
  elseif tHAZARDS[h].Name == "Flat tyre" then
    tPLAYER[p].Moves = 0
    if tPLAYER[p].SpareTyre == 0 then
      moveplayer(p, tPLAYER[p].Tile, tPLAYER[p].LastTown, true)
      tPLAYER[p].MissTurn = 2
      newturn()
    else 
      tPLAYER[p].SpareTyre = tPLAYER[p].SpareTyre - 1
      gSTAGE = 2
    end
  elseif tHAZARDS[h].Name == "Heavy traffic" then
    tPLAYER[p].Moves = math.max(tPLAYER[p].Moves-2, 0)
    gSTAGE = 2
  elseif tHAZARDS[h].Name == "Lost teddy bear" then
    if tPLAYER[p].LastTown ~= 0 then moveplayer(p, tPLAYER[p].Tile, tPLAYER[p].LastTown, true) end
    newturn()
  elseif tHAZARDS[h].Name == "MOT test date" then
    if tPLAYER[p].MOT == 0 then
      table.insert(tPLAYER[p].Convictions, "CU20: Using a vehicle in an unroadworthy condition")
    else tPLAYER[p].MOT = 0 end
    gSTAGE = 2
  elseif tHAZARDS[h].Name == "Reckless driving" then
    tPLAYER[p].Moves = tPLAYER[p].Moves * 2
    gDICE = gDICE * 2
    if gDICE > 6 then
      table.insert(tPLAYER[p].Convictions, "DD40: Reckless or dangerous driving")
      table.insert(tPLAYER[p].Convictions, "SP50: "..gDICE.."0mph in a 60mph limit near "..tTILES[getplaces("All", true)].Name)
    end
    gSTAGE = 2
  elseif tHAZARDS[h].Name == "Slow moving vehicle" then
    tPLAYER[p].Moves = math.max(tPLAYER[p].Moves-3, 0)
    gSTAGE = 2
  elseif tHAZARDS[h].Name == "Traffic jam" then
    tPLAYER[p].MissTurn = 1
    newturn()
  else gSTAGE = 2 end
  if gSTAGE == 2 and tPLAYER[gTURN].Control == "A.I." then gDELAY = math.ceil(gFPS/2) end
end

function resolveservice(p, s)
  if tSERVICES[s].Name == "Car insurance" then
    tPLAYER[p].Insurance = 1
    gSTAGE = 2
  elseif tSERVICES[s].Name == "Clear traffic" then
    tPLAYER[p].Moves = tPLAYER[p].Moves + 2
    gSTAGE = 2
  elseif tSERVICES[s].Name == "Fog lights" then
    tPLAYER[p].FogLights = 1
    gSTAGE = 2
  elseif tSERVICES[s].Name == "Lunch break" then
    tPLAYER[p].MissTurn = 1
    newturn()
  elseif tSERVICES[s].Name == "MOT certificate" then
    tPLAYER[p].MOT = 1
    gSTAGE = 2
  elseif tSERVICES[s].Name == "Shortcut" then
    gTURN = gTURN - 1
    if gTURN < 1 then gTURN = gPLAYERS end
    newturn()
  elseif tSERVICES[s].Name == "Spare tyre" then
    tPLAYER[p].SpareTyre = tPLAYER[p].SpareTyre + 1
    gSTAGE = 2
  elseif tSERVICES[s].Name == "Snow chains" then
    tPLAYER[p].SnowChains = 1
    gSTAGE = 2
  elseif tSERVICES[s].Name == "Toilet break" then
    newturn()
  else gSTAGE = 2 end
  if gSTAGE == 2 and tPLAYER[gTURN].Control == "A.I." then gDELAY = math.ceil(gFPS/2) end
end

function resolveactivity(n)
  local i, complete
  if tWEATHER[gWEATHER].Success[tPLAYER[gTURN].Activities[n].ID] == 1 then
    tPLAYER[gTURN].Activities[n].Complete = true
    playsound(3)
    complete = true
    for i = 1, #tPLAYER[gTURN].Activities, 1 do
      if tPLAYER[gTURN].Activities[i].Complete ~= true then complete = false end
    end
  else playsound(4) end
  gSTAGE = 2
  if tPLAYER[gTURN].Control == "A.I." then gDELAY = math.ceil(gFPS/2) end
  gATTEMPTED = true
  gSELECT, gSUBSELECT = 0, 0
  if complete == true then gSTAGE = 5 end
end

function playsound(i)
  if sSOUND[i] ~= nil and gSOUNDON == true then 
    love.audio.stop(sSOUND[i])
    love.audio.play(sSOUND[i])
  end
end

function ai()
  local places = getplaces("All")
  local i, j, found
  local home = false
  local home_ref = 0
  local steps, first, best, best_ref, best_first
  -- Trim my list of places to places I can still do something
  if next(tPLAYER[gTURN].Activities)~=nil and next(places)~=nil then
    for i = #places, 1, -1 do
      found = false
      for j = 1, #tPLAYER[gTURN].Activities, 1 do
        if next(tTILES[places[i]].Activity)~=nil and tPLAYER[gTURN].Activities[j].Complete ~= true then
          if tTILES[places[i]].Activity[tPLAYER[gTURN].Activities[j].ID] == 1 then   -- I can do something here
            found = true
            if places[i] == tPLAYER[gTURN].Tile then   -- This is where I am now!
              home = true
              home_ref = j
            end
          end
        end
      end
      if found == false then table.remove(places, i) end
    end
  end
  if next(tPLAYER[gTURN].Activities)~=nil and next(places)~=nil then
    if home == true then
      -- Can I do an activity where I am? If so, stop here!
      if gATTEMPTED ~= true then
        gSELECT = 1
        gSUBSELECT = home_ref
        gDELAY = gFPS*2
        gSTAGE = 4
      else newturn() end
    elseif tPLAYER[gTURN].Moves > 0 then
      -- If not then move "towards" the "best nearest" place I can!
      best, best_ref = math.huge, 0
      for i = 1, #places, 1 do
        steps, first = pathtotarget(places[i])
        if steps < best then 
          best_ref = i
          best = steps
          best_first = first
        end
      end
      if best_ref ~= 0 then
        moveplayer(gTURN, tPLAYER[gTURN].Tile, best_first, false)
      else newturn() end  
    else
      -- Or if you're out of moves
      newturn()
    end
  end
end

function police()
  local criminals = {}
  local distances = {}
  local firststep = {}
  local mostwanted, best = 0, 0
  local withinreach = false
  local tx, ty, mx, my = 0, 0, 0, 0
  local steps, first
  local i
  if tPLAYER[#tPLAYER].Moves > 0 then
    -- Make a list of criminals & distances
    if next(tPLAYER)~=nil then
      for i = 1, #tPLAYER, 1 do
        if i~=#tPLAYER and tPLAYER[i].Control~="Off" then
          tx = tx + tTILES[tPLAYER[i].Tile].X
          ty = ty + tTILES[tPLAYER[i].Tile].Y
        end
        if next(tPLAYER[i].Convictions)~=nil then 
          table.insert(criminals, i)
          steps, first = pathtotarget(tPLAYER[i].Tile)
          table.insert(distances, steps)
          table.insert(firststep, first)
          if steps <= tPLAYER[#tPLAYER].Moves then withinreach = true end
        end
      end
    end
    if next(criminals)~=nil then
      -- There's at least one criminal: who is most wanted?
      for i = 1, #criminals, 1 do
        if #tPLAYER[criminals[i]].Convictions > best and (distances[i]<=tPLAYER[#tPLAYER].Moves or withinreach == false) then
          mostwanted = i
          best = #tPLAYER[criminals[i]].Convictions
        end
      end
      -- Take the first step towards the most wanted player
      moveplayer(#tPLAYER, tPLAYER[#tPLAYER].Tile, firststep[mostwanted], false)
    else
      -- Head for the centre point between players
      mx = tx / gPLAYERS
      my = ty / gPLAYERS
      best = math.huge
      for i = 1, #tTILES, 1 do
        steps = distancebetween(mx, my, tTILES[i].X, tTILES[i].Y)
        if steps < best then
          mostwanted = i
          best = steps
        end
      end
      steps, first = pathtotarget(mostwanted)
      -- Take the first step towards the midpoint
      moveplayer(#tPLAYER, tPLAYER[#tPLAYER].Tile, first, false)
    end
  else newturn() end
end

function pathtotarget(target)
  local best = math.huge
  local best_ref = 0
  local best_first = 0
  local attempts = 0
  local i, x, first, options, moves
  local prev = tPLAYER[gTURN].Previous
  repeat
    attempts = attempts + 1
    x = tPLAYER[gTURN].Tile
    moves = 0
    repeat
      moves = moves + 1
      if next(tTILES[x].Links)~=nil then
        options = {}
        for i = 1, #tTILES[x].Links, 1 do
          if tTILES[x].Links[i] ~= prev then table.insert(options, tTILES[x].Links[i]) end
        end
        if next(options)~=nil then
          prev = x
          x = options[math.random(#options)]
          if moves == 1 then first = x end
        end
      end
    until x == target or moves == 50
    if moves < best then 
      best = moves
      best_first = first
    end
  until attempts == 1000
  return best, best_first
end

function deepcopy(orig)
  local orig_type = type(orig)
  local copy
  if orig_type == 'table' then
    copy = {}
    for orig_key, orig_value in next, orig, nil do
      copy[deepcopy(orig_key)] = deepcopy(orig_value)
    end
    setmetatable(copy, deepcopy(getmetatable(orig)))
  else copy = orig end -- number, string, boolean, etc
  return copy
end